[Setting name="Twitch OAuth Token" category="Login" description="Your OAuth token, include the 'oauth:' part. View the setup guide on the Download page for more info!" password]
string Setting_TwitchToken;

[Setting name="Twitch Nickname" category="Login" description="(Lowercase)"]
string Setting_TwitchNickname;

[Setting name="Twitch Channel" category="Login" description="(Lowercase, comma separated if multi-channel)"]
string Setting_TwitchChannel;

// -------------------------------- //
// --- METHODS FOR YOUR PLUGINS --- //
// -------------------------------- //

// CHECK METHODS TUTORIAL / DESCRIPTION IN Exports.as !!!

namespace Twitch { // begin of namespace for exports

int Register(const uint updateType) {
	int generatedId = Math::Rand(1, 2147483646);
	string generatedValue = "" + generatedId;
	while (g_pluginFilters.Exists(generatedValue)) {
		generatedId = Math::Rand(1, 2147483646);
		generatedValue = "" + generatedId;
	}
	g_pluginFilters.Set(generatedValue, updateType);
	g_pluginPointers.Set(generatedValue, g_msgCounter);
	return generatedId;
}

bool Unregister(const int pluginId) {
	string pluginKey = "" + pluginId;

	if (!g_pluginFilters.Exists(pluginKey)) {
		error("Something requested to unregister a nonexistant plugin!");
		return false;
	}

	if (!g_pluginPointers.Delete(pluginKey)) {
		error("Something went wrong unregistering a plugin from the message queue!");
		return false;
	}

	if (!g_pluginFilters.Delete(pluginKey)) {
		error("Something went wrong deleting a plugin from the ID list!");
		return false;
	} 
	
	return true;
}

array<Twitch::Message@> Fetch(const int pluginId) {
	uint requestId = g_msgCounter;

	array<Twitch::Message@> messages;
	string pluginKey = "" + pluginId;

	if (!g_pluginFilters.Exists(pluginKey)) {
		error("Something requested new messages for a nonexistant plugin!");
		return messages;
	}
	
	uint lastRequestId = uint(g_pluginPointers[pluginKey]);
	for (uint i = 0; i < g_messageQueue.Length; i++) {
		if (g_messageQueue[i].m_id >= lastRequestId) {
			if ((g_messageQueue[i].m_msg.m_messageType & uint(g_pluginFilters[pluginKey])) > 0) messages.InsertLast(g_messageQueue[i].m_msg);
		}
	}
	g_pluginPointers[pluginKey] = requestId;
	
	return messages;
}

Twitch::MessageCode SendMessage(const string &in channel, const string &in msg) {
	if (g_client is null) {
		return Twitch::MessageCode::NoConnection;
	}

	int c = 0;
	for (uint i = 0; i < g_sentMessages.Length; i++) {
		if (g_sentMessages[i] < Time::Now - 30000) c++;
	}
	g_sentMessages.RemoveRange(0,c);
	if ((g_sentMessages.Length > 0 && g_sentMessages[g_sentMessages.Length - 1] > Time::Now - 1000) || (g_sentMessages.Length > 20)) {
		return Twitch::MessageCode::RateLimited;
	} else {
		g_sentMessages.InsertLast(Time::Now);
		g_client.WriteRaw("PRIVMSG #" + channel + " :" + msg + "\n");
		return Twitch::MessageCode::Success;
	}
}

Twitch::MessageCode SendWhisper(const string &in user, const string &in msg) {
	if (g_client is null) {
		return Twitch::MessageCode::NoConnection;
	}
	// same checking methodology as normal messages
	int c = 0;
	for (uint i = 0; i < g_sentWhispers.Length; i++) {
		if (g_sentWhispers[i] < Time::Now - 60000) c++;
	}
	g_sentWhispers.RemoveRange(0,c);
	if ((g_sentWhispers.Length > 0 && g_sentWhispers[g_sentWhispers.Length - 1] > Time::Now - 1000) || (g_sentWhispers.Length > 100)) {
		return Twitch::MessageCode::RateLimited;
	} else {
		g_sentWhispers.InsertLast(Time::Now);
		if (g_channels.Length < 1) return Twitch::MessageCode::NoConnection;
		g_client.WriteRaw("PRIVMSG #" + g_channels[0].m_name + " :/w " + user + " " + msg + "\n");
		return Twitch::MessageCode::Success;
	}
}

array<Twitch::ChannelState@> GetJoinedChannels() {
	array<Twitch::ChannelState@> channels;
	for (uint i = 0; i < g_channels.Length; i++) {
		if (g_channels[i].m_checked && g_channels[i].m_exists) channels.InsertLast(g_channels[i]);
	}
	return channels;
}

bool ChannelsJoined() {
	return g_channelsJoined;
}

array<Twitch::ChannelState@> GetAllChannels() {
	return g_channels;
}

bool ChannelsAdded() {
	return g_channelsAdded;
}

string GetUsername() {
	return Setting_TwitchNickname;
}

} // end of namespace "Twitch"

// --------------------------------------- //
// --- END OF METHODS FOR YOUR PLUGINS --- //
// --------------------------------------- //

// --- Globals

Net::SecureSocket@ g_client;

array<Twitch::ChannelState@> g_channels;
bool g_channelsAdded = false;
bool g_channelsJoined = false;

uint64[] g_sentMessages;
uint64[] g_sentWhispers;

array<StoredTwitchMessage@> g_messageQueue;
dictionary g_pluginPointers;
dictionary g_pluginFilters;
uint g_msgCounter;

// --- Networking and login

void Disconnect() {
	if (g_client is null) {
		return;
	}
	g_client.Close();
	@g_client = null;
	g_channelsAdded = false;
	g_channelsJoined = false;
}

bool SetupSocket() {
	if (g_client !is null) {
		Disconnect();
	}

	// Initial Socket Setup
	uint connectionAttempts = 0;
	while (connectionAttempts < 5) {
		connectionAttempts++;

		@g_client = Net::SecureSocket();
		if (!g_client.Connect("irc.chat.twitch.tv", 6697)) {
			error("Socket setup to Twitch IRC server failed, retrying...");
			sleep(5000);
			continue;
		}

		uint64 timeoutDuration = 10000;
		uint64 nextTry = Time::Now + timeoutDuration;
		bool connectionTimedOut = false;
		while (g_client.Connecting()) {
			if (Time::Now > nextTry) {
				warn("Initial Twitch connection timed out (" + timeoutDuration / 1000 + "s), retrying...");
				connectionTimedOut = true;
				break;
			}

			yield();
		}

		if (connectionTimedOut) {
			// abandon socket
			continue;
		}

		// Socket could connect and exited Connecting state before timeout, LETS GO
		return true;
	}

	error("Maximum amount of Twitch connection attempts reached.");
	return false;
}

bool SetupIRC() {

	// IRC setup (including Login)
	uint loginAttempts = 0;
	while (loginAttempts < 5) {
		loginAttempts++;

		if (!SetupSocket()) {
			return false;
		}

		if (!g_client.WriteRaw(
			"PASS " + Setting_TwitchToken + "\n"
			"NICK " + Setting_TwitchNickname + "\n"
		)) {
			// WeirdChamp if this happens
			error("Couldn't send login data to Twitch!");
			continue;
		}

		uint64 timeoutDuration = uint64(Math::Pow(2,loginAttempts)) * 1000;
		uint64 nextTry = Time::Now + timeoutDuration;
		while (Time::Now < nextTry) {

			string line;
			if (g_client.ReadLine(line)) {
				line = line.Trim();

				auto msg = IRC::Message(line);
				if (msg.m_command == "001") {
					// Twitch replied successfully!
					return true;
				} else {
					// Connection works but stuff like credentials could be wrong, important to stop here
					throw("Twitch auth error: " + line);
				}
			}
			
			yield();
		}
		
		warn("Retrying login to twitch (timeout " + timeoutDuration / 1000 + "s)...");
	}

	error("Maximum amount of connection / login attempts reached.");
	return false;
}

void JoinChannels() {
	if (g_client is null || !g_client.CanWrite()) {
		throw("Tried joining channels when connection isnt established.");
	}

	g_client.WriteRaw("CAP REQ :twitch.tv/tags twitch.tv/commands\n");

	// wait for CAP * ACK
	while (true) {
		string line;
		if (g_client.ReadLine(line)) {
			line = line.Trim();

			auto msg = IRC::Message(line);
			if (msg.m_command == "CAP") {
				if (msg.m_params[1] == "ACK") {
					break;
				} else {
					throw("Required capabilities not accepted by Twitch!");
				}
			}
		}

		yield();
	}

	g_channelsAdded = false;
	g_channelsJoined = false;
	g_channels.RemoveRange(0, g_channels.Length);
	string[] channels = Setting_TwitchChannel.Replace(" ","").ToLower().Split(",");

	for (uint i = 0; i < channels.Length; i++) {
		Twitch::ChannelState@ channel = Twitch::ChannelState(channels[i]);
		g_client.WriteRaw("JOIN #" + channel.m_name + "\n");
		g_channels.InsertLast(channel);
	}

	g_channelsAdded = true;
	startnew(ChannelNotFoundCheck);
	UI::ShowNotification("\\$94f" + Icons::Check + " Connected to Twitch IRC!", "Joining stream chats...", 5000);
	trace("Connection successful!");
}

void ChannelNotFoundCheck() {
	uint64 joinTimeout = Time::Now + 15000;
	while (true) {
		yield();

		bool joined = true;
		for (uint i = 0; i < g_channels.Length; i++) {
			if (!g_channels[i].m_checked) joined = false;
		}

		if (joined) {
			break;
		}

		if (Time::Now > joinTimeout) {
			for (uint i = 0; i < g_channels.Length; i++) {
				if (!g_channels[i].m_checked) {
					if (g_channels[i].m_exists) {
						error("Channel " + g_channels[i].m_name + " exists but reuqesting ROOMSTATE / USERSTATE timed out!");
						if (g_channels[i].m_userstate is null) @g_channels[i].m_userstate = Twitch::Message();
						if (g_channels[i].m_roomstate is null) @g_channels[i].m_roomstate = Twitch::Message();
					} else {
						WarnUser("\\$f55The channel " + g_channels[i].m_name + " does not exist or the connection failed!");
					}
					g_channels[i].m_checked = true;
				}
			}
			break;
		}
	}
	g_channelsJoined = true;
}

// --- Twitch specific structures and data

Twitch::Message@ Twitchify(const IRC::Message@ &in msg, const Twitch::MessageType messageType) {
	auto message = Twitch::Message();
	message.m_messageType = messageType;
	if (msg.m_params.Length > 0) message.m_channel = msg.m_params[0].SubStr(1,msg.m_params[0].Length - 1);
	if (msg.m_params.Length > 1) message.m_text = msg.m_params[1];
	message.m_username = msg.m_prefix.m_origin;
	msg.m_tags.Get("display-name", message.m_username);
	msg.m_tags.Get("tmi-sent-ts", message.m_timestamp);
	message.m_tags = msg.m_tags;
	return message;
}

class StoredTwitchMessage {
	Twitch::Message m_msg;
	uint64 m_storeTime;
	uint m_id;

	StoredTwitchMessage(const Twitch::Message &in msg) {
		m_msg = msg;
		m_storeTime = Time::Now;
		m_id = g_msgCounter;
		g_msgCounter++;
	}
}

StoredTwitchMessage@ ConvertMessage(const IRC::Message@ &in msg, const Twitch::MessageType messageType) {
	return StoredTwitchMessage(Twitchify(msg, messageType));
}

void UpdateTwitch() {
	if (g_client is null || g_client.Available() == 0) {
		return;
	}

	string line;
	if (g_client.ReadLine(line)) {
		line = line.Trim();

		auto msg = IRC::Message(line);
		if (msg.m_command == "PRIVMSG") {
			g_messageQueue.InsertLast(ConvertMessage(msg, Twitch::MessageType::ChatMessage));
		} else if (msg.m_command == "USERNOTICE") {
			g_messageQueue.InsertLast(ConvertMessage(msg, Twitch::MessageType::UserNotice));
		} else if (msg.m_command == "USERSTATE") {
			Twitch::Message@ message = Twitchify(msg, Twitch::MessageType::UserState);
			for (uint i = 0; i < g_channels.Length; i++) {
				if (message.m_channel == g_channels[i].m_name) {
					@g_channels[i].m_userstate = message;
					if (!g_channels[i].m_checked) {
						g_channels[i].m_exists = true;
						if (g_channels[i].m_roomstate !is null) g_channels[i].m_checked = true;
					}
					break;
				}
			}
			g_messageQueue.InsertLast(StoredTwitchMessage(message));
		} else if (msg.m_command == "WHISPER") {
			g_messageQueue.InsertLast(ConvertMessage(msg, Twitch::MessageType::Whisper));
		} else if (msg.m_command == "CLEARCHAT") {
			g_messageQueue.InsertLast(ConvertMessage(msg, Twitch::MessageType::ClearChat));
		} else if (msg.m_command == "CLEARMSG") {
			g_messageQueue.InsertLast(ConvertMessage(msg, Twitch::MessageType::ClearMessage));
		} else if (msg.m_command == "HOSTTARGET") {
			g_messageQueue.InsertLast(ConvertMessage(msg, Twitch::MessageType::HostTarget));
		} else if (msg.m_command == "NOTICE") {
			if (string(msg.m_tags["msg_id"]) == "msg_channel_suspended") {
				error("The channel " + msg.m_params[1] + "is either suspended or deleted!");
				for (uint i = 0; i < g_channels.Length; i++) {
					if (msg.m_params[1] == "#" + g_channels[i].m_name) {
						g_channels[i].m_exists = false;
						g_channels[i].m_checked = true;
						break;
					}
				}
			}
			g_messageQueue.InsertLast(ConvertMessage(msg, Twitch::MessageType::Notice));
		} else if (msg.m_command == "JOIN") {
			for (uint i = 0; i < g_channels.Length; i++) {
				if (msg.m_params[0] == "#" + g_channels[i].m_name) {
					trace("Joined channel " + g_channels[i].m_name);
					g_channels[i].m_exists = true;
				}
			}
			g_messageQueue.InsertLast(ConvertMessage(msg, Twitch::MessageType::Join));
		} else if (msg.m_command == "RECONNECT") {
			Disconnect();
		} else if (msg.m_command == "ROOMSTATE") {
			Twitch::Message@ message = Twitchify(msg, Twitch::MessageType::RoomState);
			for (uint i = 0; i < g_channels.Length; i++) {
				if (message.m_channel == g_channels[i].m_name) {
					@g_channels[i].m_roomstate = message;
					if (!g_channels[i].m_checked) {
						g_channels[i].m_exists = true;
						if (g_channels[i].m_userstate !is null) g_channels[i].m_checked = true;
					}
					break;
				}
			}
			g_messageQueue.InsertLast(StoredTwitchMessage(message));
		} else if (msg.m_command == "GLOBALUSERSTATE") {
			g_messageQueue.InsertLast(ConvertMessage(msg, Twitch::MessageType::GlobalUserState));
		} else if (msg.m_command == "PING") {
			g_client.WriteRaw("PONG :" + msg.m_params[0] + "\n");
		} else if (msg.m_command == "366") {
			for (uint i = 0; i < g_channels.Length; i++) {
				if (msg.m_params[0] == "#" + g_channels[i].m_name) {
					if (g_channels[i].m_userstate is null || g_channels[i].m_userstate is null) {
						error("Received 366 for channel " + g_channels[i].m_name + " even though there's no USERSTATE / ROOMSTATE yet!");
						if (g_channels[i].m_userstate is null) @g_channels[i].m_userstate = Twitch::Message();
						if (g_channels[i].m_roomstate is null) @g_channels[i].m_roomstate = Twitch::Message();
					}
					g_channels[i].m_exists = true;
					g_channels[i].m_checked = true;
					break;
				}
			}
		} else if (msg.m_command != "353") {
			trace("[IRC] " + line);
		}
	}
}

// --- Misc stuff

void WarnUser(const string &in msg) {
	UI::ShowNotification(msg, 20000);
	print(msg);
}

void ResetRateLimits() {
	g_sentMessages.RemoveRange(0, g_sentMessages.Length);
	g_sentWhispers.RemoveRange(0, g_sentWhispers.Length);
	g_sentMessages.InsertLast(0);
	g_sentWhispers.InsertLast(0);
}

// --- Main and Openplanet Plugin functions

void Main() {
	if (Setting_TwitchToken == "" && Setting_TwitchNickname == "" && Setting_TwitchChannel == "") {
		WarnUser("\\f55The Twitch base isn't set up yet! Please follow the guide on the Plugin download page.");
		return;
	}

	if (!Setting_TwitchToken.StartsWith("oauth:")) {
		WarnUser("\\$f55No / invalid Twitch token set. Set one in the settings and reload the plugin!\nAlso, remember to include the\"oauth:\"part in front.");
		return;
	}

	if (Setting_TwitchNickname == "") {
		WarnUser("\\$f55No Twitch nickname set. Please set one in the settings and reload the plugin!");
		return;
	}

	if (Setting_TwitchChannel == "") {
		WarnUser("\\$f55No Twitch channel set. Please set one in the settings and reload the plugin!");
		return;
	}

	while (true) {
		if (g_client is null || !g_client.CanWrite()) {
			SetupTwitch();
		}
		LoopQueue();
		UpdateTwitch();
		yield();
	}
}

void OnDisabled() {
	Disconnect();
}

// --- Control flow

void SetupTwitch() {
	UI::ShowNotification("Connecting to Twitch...");
	if (SetupIRC()) {
		JoinChannels();
	} else {
		throw("Could not setup IRC, exiting.");
	}
	ResetRateLimits();
}

void LoopQueue() {
	uint64 deletionTime = Time::Now - 1000;
	uint removeCount = 0;
	for (uint i = 0; i < g_messageQueue.Length; i++) {
		if (g_messageQueue[i].m_storeTime < deletionTime) {
			removeCount++;
		} else break;
	}
	g_messageQueue.RemoveRange(0, removeCount);
}